from pathlib import Path

import typer
import pkg_resources

app = typer.Typer(
    name="linc CLI",
    help="Converts raw lidar files (Licel format) to NC with a configuration file",
)


@app.command(no_args_is_help=True)
def convert(
    source_path: list[Path] = typer.Argument(..., dir_okay=False, readable=True),
    output_path: Path = typer.Option(..., "--output", "-o", file_okay=True),
    config_file: Path = typer.Option(
        None,
        "--config",
        "-c",
    ),
    legacy: bool = typer.Option(False, "--legacy", "-l"),
):
    ...
    # (source_path, output_file=output_path, config_file=config_file)


def version_callback(value: bool):
    if value:
        version = pkg_resources.get_distribution("fastapi-cli").version
        typer.echo(f"fastapi-cli, version {version}")
        raise typer.Exit()


@app.callback()
def main(
    version: bool = typer.Option(
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Show the Manage FastAPI version information.",
    )
):
    ...



from pathlib import Path
from pdb import set_trace

import numpy as np
import pandas as pd

from .convertion import convert_to_physical_units
from .models import DataFileU32, DataFile
from .config import Config
from .parse.header import parse_header
from .parse.dataset import parse_dataset
from .parse.file import read_file_header_dataset


def read_file(file_path: str | Path, config: Config | None = None) -> DataFile:
    _p = Path(file_path)
    h, d = read_file_header_dataset(_p)
    header = parse_header(h.split(b"\r\n"))
    dataset = parse_dataset(d, header=header)

    file_u32 = DataFileU32(header=header, dataset=dataset)
    file = convert_to_physical_units(file_u32)

    if config is not None:
        new_dataset = replace_with_names(file.dataset, config)
        file = DataFile(header=header, dataset=new_dataset)

    return file


def replace_with_names(dataset: pd.DataFrame | pd.Series, config: Config) -> pd.DataFrame | pd.Series:
    final_columns = list(map(lambda ch: ch.link_to, config.lidar.channels))
    # set_trace()

    if config.options.channel_not_present_all_ok:
        select_columns = list(filter(lambda ch: ch in dataset.columns, final_columns))
    else:
        select_columns = final_columns

    dataset = dataset[select_columns]
    

    for channel in config.lidar.channels:
        if channel.link_to in dataset.columns:
            if isinstance(dataset, pd.DataFrame):
                dataset = dataset.rename(columns={channel.link_to: channel.name})
            else:
                dataset = dataset.rename({channel.link_to: channel.name})
        else:
            if isinstance(dataset, pd.DataFrame):
                dataset.loc[:, channel.name] = np.nan
            else:
                dataset[:] = np.nan

    return dataset



import numpy as np
import pandas as pd
import numpy.typing as npt

from linc.utils import device_id_to_str

from ..models import Header


def parse_dataset(dataset: bytes, header: Header) -> pd.DataFrame:
    dt = np.dtype(np.uint32)
    dt = dt.newbyteorder("<")

    max_length = max(map(lambda x: x.bins, header.channels))
    parsed = np.full((len(header.channels), max_length), fill_value=np.nan)
    _parsed = pd.DataFrame()

    for channel in header.channels:
        bytes_size = channel.bins * 4
        current_array = np.frombuffer(dataset, dtype=dt, count=channel.bins)
        _parsed[device_id_to_str(channel.device_id)] = np.pad(
            current_array, (0, max_length - current_array.shape[0]), "empty"
        )
        # parsed[idx, : current_array.shape[0]] = current_array

        *_, dataset = dataset[bytes_size:].partition(b"\r\n")
        # print("parsed: ")
        # print(parsed)
        # print(before, symbol)

    # set_trace()

    return _parsed



# type: ignore
from datetime import datetime
from pdb import set_trace

from ..models import Channel, DeviceId, Header, Laser
from ..utils import safe_get_list
from .constants import LICEL_DATE_FORMAT


def parse_header(header_list: list[bytes]) -> Header:
    l2 = header_list[1].split()
    l3 = header_list[2].split()
    l_channels = header_list[3:]

    lasers = parse_lasers(l3)
    channels = parse_channels(l_channels)

    return Header(
        filename=header_list[0],
        location=l2[0],
        start_date=datetime.strptime(
            f"{b' '.join(l2[1: 3]).decode()}", LICEL_DATE_FORMAT
        ),
        stop_date=datetime.strptime(
            f"{b' '.join(l2[3: 5]).decode()}", LICEL_DATE_FORMAT
        ),
        altitude=l2[5],
        longitude=l2[6],
        latitude=l2[7],
        zenit_angle=l2[8],
        azimuth_angle=safe_get_list(l2, 9, None),
        n_datasets=l3[4],
        lasers=lasers,
        channels=channels,
    )


def parse_lasers(lasers_line: list[bytes]) -> tuple[Laser]:

    if len(lasers_line) < 5:
        raise ValueError("Cannot parse lasers line from file")

    laser1 = Laser(shots=lasers_line[0], frecuency=lasers_line[1])
    laser2 = Laser(shots=lasers_line[2], frecuency=lasers_line[3])

    if len(lasers_line) == 5:
        return laser1, laser2

    laser3 = Laser(shots=lasers_line[5], frecuency=lasers_line[6])

    if len(lasers_line) == 7:
        return laser1, laser2, laser3
    elif len(lasers_line) == 9:
        laser4 = Laser(shots=lasers_line[7], frecuency=lasers_line[8])
        return laser1, laser2, laser3, laser4
    else:
        raise NotImplementedError("Unknown number of lasers tried to parse")


def parse_channels(channels_lines: list[bytes]) -> tuple[Channel]:
    channels: list[Channel] = []
    for line in channels_lines:
        pl = line.split()  # Properties list

        channel_physics = pl[7].split(b".")

        channels.append(
            Channel(
                active=pl[0],
                type=pl[1],
                laser=pl[2],
                bins=pl[3],
                laser_polarization=pl[4].decode(),
                ptm_voltage=pl[5],
                binwidth=pl[6],
                wavelength=channel_physics[0],
                polarization=channel_physics[1].decode(),
                adc_bits=pl[12],
                shots=pl[13],
                dc_dr=pl[14],
                device_id=parse_device_id(pl[15]),
            )
        )

    return tuple(channels)  # Tuple used to guarantee integrity


def parse_device_id(bytes_id: bytes) -> DeviceId:
    device_id = bytes_id.decode()
    sep_index = 3 if device_id.startswith("S2") else 2

    # set_trace()
    return DeviceId(type=device_id[:sep_index], number=device_id[sep_index:])



import pkg_resources
from typing import TypeVar, Iterator

from .models import DeviceId, MeasurementTypeEnum


U = TypeVar("U")
T = TypeVar("T", str, bytes)


def split_seq(seq: list[T], sep: T) -> Iterator[list[T]]:
    start = 0
    while start < len(seq):
        try:
            stop = start + seq[start:].index(sep)
            yield seq[start:stop]
            start = stop + 1
        except ValueError:
            yield seq[start:]
            break


def safe_get_list(l: list[U], idx: int, default: U | None) -> U | None:
    try:
        return l[idx]
    except IndexError:
        return default


def to_acquisition_type_string(type: MeasurementTypeEnum) -> str:
    match type:
        case MeasurementTypeEnum.ANALOG:
            return "a"
        case MeasurementTypeEnum.ANALOG_SQUARED:
            return "A"
        case MeasurementTypeEnum.PHOTONCOUNTING:
            return "p"
        case MeasurementTypeEnum.PHOTONCOUNTING_SQUARED:
            return "P"
        case _:
            raise ValueError("Input type not supported")


def device_id_to_str(device_id: DeviceId) -> str:
    return f"{device_id.type}{device_id.number}"




from pdb import set_trace
import pkg_resources

from .write.writer_nc import write_nc_legacy
from .config.read import get_config

__version__ = pkg_resources.get_distribution("linc").version

__all__ = [
    "write_nc_legacy",
    "get_config"
]


from ..config import Config
from ..models import Channel, Header
from ..utils import device_id_to_str, safe_get_list, to_acquisition_type_string


def compare_joinable_dataset(header1: Header, header2: Header) -> None:
    assert (
        header1.channels == header2.channels
    ), "While parsing a same dataset, channels most remain identical"

    assert (
        header1.location == header2.location
    ), "While parsing a same dataset, channels most remain identical"



def get_merged_channel_config(channel: Channel, config: Config) -> Channel:
    channel_as_str = device_id_to_str(channel.device_id)
    channel_config = safe_get_list(
        list(filter(lambda c: c.link_to == channel_as_str, config.lidar.channels)),
        0,
        None,
    )

    if channel_config is None:
        return channel

    return Channel(**channel.copy(update=channel_config).dict())  # type: ignore


def format_channel(channel: Channel, format: str) -> str:
    format = format.replace(r"%w", str(channel.wavelength))
    format = format.replace(r"%p", channel.polarization.value)
    format = format.replace(r"%a", to_acquisition_type_string(channel.device_id.type))
    format = format.replace(
        r"%i", f"{channel.device_id.type}{channel.device_id.number}"
    )

    return format



import time
import warnings
import pkg_resources
from pathlib import Path
from typing import Any
from collections.abc import Iterable

import numpy as np
from cftime import date2num

from ..models import DataFile
from ..reader import read_file
from ..config import Config
from ..parse.utils import parse_date_from_filename

with warnings.catch_warnings():
    warnings.filterwarnings("ignore")
    from netCDF4 import Dataset  # type: ignore


def write_nc_legacy(
    files: Iterable[Path | str],
    output_file: Path | str,
    config: Config,
) -> None:
    _f = sorted(list(files))

    nc = Dataset(output_file, "w")
    nc.history = "Created " + time.ctime(time.time())
    nc.generated_with = str(pkg_resources.get_distribution("linc").egg_name())
    
    bin_width = config.lidar.bin_width
    bin_count = config.lidar.bin_count
    channel_array = np.array(list(map(lambda c: c.name, config.lidar.channels)))
    first_time = parse_date_from_filename(Path(_f[0]).name)

    time_dim = nc.createDimension("time", None)
    range_dim = nc.createDimension("range", config.lidar.bin_count)
    channel_dim = nc.createDimension("channel", channel_array.shape[0])

    time_var = nc.createVariable("time", "i4", ("time",), compression="zlib")
    time_var.units = f"milliseconds since {first_time.isoformat().replace('T', ' ')}"
    time_var.calendar = "proleptic_gregorian"

    range_var = nc.createVariable(
        "range", "f4", ("range",), compression="zlib"
    )
    range_var[:] = np.arange(
        bin_width, bin_width * (bin_count + 1), bin_width
    )
    
    channel_var = nc.createVariable(
        "channel", "S8", ("channel",)
    )
    channel_var[:] = channel_array
    
    signal_vars = create_signal_variables(nc, config)
    
    for idx_f, iter_file in enumerate(_f):
        current_file = read_file(iter_file, config = config)
        
        write_signal_vars(current_file, time_var, signal_vars, idx_f)

    write_attrs(nc, config)
    

    nc.close()
    
def write_signal_vars(current_file: DataFile, time_var: Any ,signal_vars: list[Any], index: int) -> None:
    time_var[index] = date2num(
            current_file.header.start_date, units=time_var.units, calendar = time_var.calendar
        )
    for (channel_str, channel_var) in signal_vars:
            channel_var[index, :] = current_file.dataset[channel_str].values  # type: ignore
    

def write_attrs(nc: Any, config: Config) -> None:
    for k, v in config.lidar.attrs.items():
        set_or_create_attr(nc, attr_name=k, attr_value=v)


def create_signal_variables(nc: Any, config: Config) -> list[Any]:
    channels_vars: list[tuple[str, Any]] = []
    
    for channel in list(map(lambda x: x.name, config.lidar.channels)):
        channel_str = channel

        try:
            signal_var = nc.createVariable(
                f"signal_{channel_str}",
                "f4",
                ("time", "range"),
                compression="zlib",
            )
        except:
            raise ValueError(f"problem creating variable: {channel_str}")

        channels_vars.append((channel_str, signal_var))
    return channels_vars

def set_or_create_attr(var, attr_name, attr_value):
    if attr_name in var.ncattrs(): 
        var.setncattr(attr_name, attr_value)
        return
    try:
        var.UnusedNameAttribute = attr_value
        var.renameAttribute("UnusedNameAttribute", attr_name)
    except TypeError:
        raise TypeError(f"Type of attribute {attr_name} is {type(attr_value)} which cannot be written in netCDF")
    return



import numpy as np
import pandas as pd


from .models import Channel, DataFile, DataFileU32, MeasurementTypeEnum
from .utils import device_id_to_str

SQUARED_CHANNELS = [
    MeasurementTypeEnum.ANALOG_SQUARED,
    MeasurementTypeEnum.PHOTONCOUNTING_SQUARED,
]


def convert_to_physical_units(data_file: DataFileU32) -> DataFile:
    final_dataset = data_file.dataset.copy()

    squared_channels = list(
        filter(
            lambda c: c.device_id.type in SQUARED_CHANNELS,
            data_file.header.channels,
        )
    )

    measurement_channels = list(
        filter(
            lambda c: c.device_id.type not in SQUARED_CHANNELS,
            data_file.header.channels,
        )
    )

    # First iterate over normal channels
    for channel in measurement_channels:
        channel_str = device_id_to_str(channel.device_id)
        final_dataset[channel_str] = _to_physical(channel, final_dataset[channel_str])

    for channel in squared_channels:
        channel_str = device_id_to_str(channel.device_id)
        final_dataset[channel_str] = _to_standard_deviation(
            channel,
            final_dataset[channel_str],
            # data_file.dataset[measurement_dataset_idx].astype(np.float64),
        )

    return DataFile(header=data_file.header, dataset=final_dataset)


def _to_physical(channel: Channel, dataset: pd.Series) -> pd.Series:
    match channel.device_id.type:
        case MeasurementTypeEnum.ANALOG:
            _d = (1000 * dataset * channel.dc_dr) / (
                channel.shots * (2**channel.adc_bits - 1)
            )
        case MeasurementTypeEnum.PHOTONCOUNTING:
            _d = (dataset) / (
                channel.shots
                * 1e6
                * channel.binwidth  # Microseconds conversion output in MHz
            )
        case MeasurementTypeEnum.POWERMETER:
            _d = dataset / channel.shots  # Just apply normalization
        case MeasurementTypeEnum.POWERMETER_PHOTODIODE:
            _d = dataset / channel.shots  # Just apply normalization
        case default:
            # _d = (dataset) / (channel.shots * (2**channel.adc_bits - 1))
            raise ValueError(f"Channel not handled: {default}")
    return _d


def _to_standard_deviation(
    channel: Channel,
    dataset: pd.Series,
    # measurement_dataset: npt.NDArray[np.float64],
) -> pd.Series:
    match channel.device_id.type:
        case MeasurementTypeEnum.ANALOG_SQUARED:
            _s = (
                (1000 * dataset * channel.dc_dr)
                / (channel.shots * (2**channel.adc_bits - 1))
                / np.sqrt(channel.shots - 1)
            )
        case MeasurementTypeEnum.PHOTONCOUNTING_SQUARED:
            _s = (
                (1000 * dataset * channel.dc_dr)
                / (channel.shots * (2**channel.adc_bits - 1))
                / np.sqrt(channel.shots)
            )
        case default:
            raise ValueError(f"Channel not handled: {default}")

    _s[channel.bins:] = np.nan # Remove other non-square-mean

    return _s



from typing import Any

from pydantic import BaseModel, Field


DEFAULT_ATTRS: dict[str, Any] = {"generated_by": "linc-python"}


class Options(BaseModel):
    channel_not_present_all_ok: bool = False

class LidarChannels(BaseModel):
    name: str
    link_to: str # The device ID


class LidarConfig(BaseModel):
    bin_width: float 
    bin_count: int
    attrs: dict[str, Any] = DEFAULT_ATTRS
    channels: list[LidarChannels] = Field(default_factory=list)


class Config(BaseModel):
    lidar: LidarConfig = Field()
    options: Options = Field()


"""
DEFAULT_CONFIG: Config = {
    "lidar": {
        "attrs": {"converter": "linc"},
        "channels": [
            # {"wavelength": 532, "link_to": "BT0"},
            # {"wavelength": 532, "link_to": "S2A0"},
            # {"wavelength": 532, "link_to": "BT1"},
            # {"wavelength": 532, "link_to": "S2A1"},
        ],
    },
    "config": {
        "include_undefined_channels": True,
        "default_channel_name_format": r"%wx%p%a",
    },
}
"""



# from typing import Any, TypedDict

# class SubConfig(TypedDict):
#     include_undefined_channels: bool
#     default_channel_name_format: str

# class LidarChannels(TypedDict):
#     wavelength: int
#     link_to: str


# class LidarConfig(TypedDict):
#     attrs: dict[Any, Any]
#     channels: list[LidarChannels]


# class Config(TypedDict):
#     lidar: LidarConfig
#     config: SubConfig


# __doc__ = r"""
# Configuration file.

# config.default_channel_name_format Tokens:
# %w Wavelength (3-4 digits integer)
# %p Polarization state  (o, p, s, r, l)
# %a Adquisition mode (a, p, A, P)  A and P is for squared datasets

# """


from enum import Enum
from datetime import datetime

import pandas as pd
from pydantic import BaseModel, validator


class LaserPolarizationEnum(str, Enum):
    NONE = "0"
    VERTICAL = "1"
    HORIZONTAL = "2"
    RIGHT_CIRCULAR = "3"
    LEFT_CIRCULAR = "4"


class PolarizationEnum(str, Enum):
    NONE = "o"
    PARALLEL = "p"
    CROSSED = "s"
    RIGHT_CIRCULAR = "r"
    LEFT_CIRCULAR = "l"


class MeasurementTypeEnum(str, Enum):
    ANALOG = "BT"
    PHOTONCOUNTING = "BC"
    ANALOG_SQUARED = "S2A"
    PHOTONCOUNTING_SQUARED = "S2P"
    POWERMETER_PHOTODIODE = "PD"
    POWERMETER = "PM"


class FileTypeEnum(str, Enum):
    RAW_SIGNAL = "RS"
    DARK_CURRENT = "DC"
    TELECOVER = "TC"
    DESPOLARIZATION = "DP"


class Laser(BaseModel):
    shots: int
    frecuency: int


class DeviceId(BaseModel):
    number: int
    type: MeasurementTypeEnum


class Channel(BaseModel):
    active: bool
    laser: int
    bins: int
    laser_polarization: LaserPolarizationEnum = LaserPolarizationEnum.NONE
    ptm_voltage: int
    binwidth: float
    wavelength: int
    polarization: PolarizationEnum = PolarizationEnum.NONE
    adc_bits: int
    shots: int
    dc_dr: float  # Discriminator level () or data range
    device_id: DeviceId

    class Config:
        allow_mutation = False


class Header(BaseModel):
    filename: str
    location: str
    start_date: datetime
    stop_date: datetime
    altitude: float
    longitude: float
    latitude: float
    zenit_angle: float
    azimuth_angle: float | None = None
    n_datasets: int
    lasers: list["Laser"]
    channels: list["Channel"]

    class Config:
        anystr_strip_whitespace = True  # remove trailing whitespace
        allow_mutation = False

    # TODO: write custom validator for channels with associated lasers
    # @validator('channels')
    # def check_laser_on_channel(cls, v, _):
    #     return v


class DataFileU32(BaseModel):
    header: Header
    dataset: pd.DataFrame | pd.Series

    # Validator not valid anymore since config files ...
    # @validator("dataset")
    # def dataset_length_must_match_header(cls, v, values):
    #     _header: Header = values["header"]

    #     # Header shape: (datasets x max bin of single profile)
    #     header_shape = (
    #         max(map(lambda x: x.bins, _header.channels)),
    #         _header.n_datasets,
    #     )
    #     if v.shape != header_shape:
    #         raise ValueError(
    #             f"Header shape: {header_shape} does not match with dataset shape {v.shape}"
    #         )

    #     return v

    class Config:
        arbitrary_types_allowed = True
        allow_mutation = False


class DataFile(DataFileU32):
    dataset: pd.DataFrame | pd.Series



from pdb import set_trace
from datetime import datetime



def parse_date_from_filename(filename: str, /) -> datetime:
    date = f"{filename[2:4]}-{int(filename[4], base = 16):02d}-{filename[5:7]}"
    hour = f"{filename[7:9]}:{filename[10:12]}:{filename[12:14]}"

    return datetime.strptime(f"{date}T{hour}", r"%y-%m-%dT%H:%M:%S")
    
    


from functools import lru_cache
from pathlib import Path

from pydantic import BaseModel

from tomli import load
from .models import Config

FALLBACK_CONFIG = Path(__file__).parent / "default-config.toml"
CURRENT_DIR_CONFIG = Path("./lidar-conf.toml")


@lru_cache
def get_config(file: Path | str | None = None, use_default: bool = False) -> Config:

    if file is not None:
        _cf = Path(file)
        if _cf.exists():
            config_path = _cf
        elif CURRENT_DIR_CONFIG.exists():
            config_path = CURRENT_DIR_CONFIG
        else:
            if use_default:
                config_path = FALLBACK_CONFIG
            else:
                raise FileNotFoundError()
    else:
        if use_default:
                config_path = FALLBACK_CONFIG
        else:
            raise FileNotFoundError()

    with open(config_path, "rb") as f:
        settings_dict = load(f)

    return Config(**settings_dict)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['linc', 'linc.config', 'linc.parse', 'linc.write']

package_data = \
{'': ['*']}

install_requires = \
['netCDF4>=1.6.2,<2.0.0',
 'numpy>=1.23.4,<2.0.0',
 'pydantic>=1.10.2,<2.0.0',
 'tomli>=2.0.1,<3.0.0',
 'typer[all]>=0.6.1,<0.7.0']

entry_points = \
{'console_scripts': ['linc = linc.cli:app']}

setup_kwargs = {
    'name': 'linc',
    'version': '1.0.0',
    'description': 'A package to handle LICEL Binary format',
    'long_description': '# Linc\n---\nLinc is a fast way to convert [Licel Raw Format](https://licel.com/raw_data_format.html) into netCDF4 to handle data easier.\n',
    'author': 'Juan Diego',
    'author_email': 'jdlar@eafit.edu.co',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<3.12',
}


setup(**setup_kwargs)



from pathlib import Path

def read_file_header_dataset(file_path: Path | str) -> tuple[bytes, bytes]:
    _p = Path(file_path)

    with open(_p, "rb") as f:
        header, _, dataset = f.read().partition(b"\r\n\r\n")

    return header, dataset
